from __future__ import annotations

import gzip
from abc import ABCMeta, abstractmethod
from io import IOBase
from typing import Any, Generic, Iterable, Type, overload

import requests
from typing_extensions import Literal, get_args

from unipressed.dataset.search import Search
from unipressed.dataset.type_vars import (
    FieldsType,
    FormatType,
    JsonResultType,
    QueryType,
)


class DatasetClient(
    Generic[QueryType, JsonResultType, FieldsType, FormatType],
    metaclass=ABCMeta,
):
    """
    The base class for all UniProt dataset clients. All methods documented here are available in any of the subclasses. This is a static class that you will never need to instantiate.
    """

    @classmethod
    def _id_field(cls, record: JsonResultType) -> str:
        """
        Given a record, extracts the accession/ID field from it.
        """
        return record["id"]

    @classmethod
    @abstractmethod
    def name(cls) -> str:
        ...

    @classmethod
    def search(
        cls,
        query: QueryType,
        format: FormatType | Literal["json"] = "json",
        fields: Iterable[FieldsType] | None = None,
        size: int = 500,
        include_isoform: bool = True,
    ) -> Search[QueryType, JsonResultType, FieldsType, FormatType]:
        """
        Creates an object that can be used to perform a search query over this dataset.
        Refer to the [unipressed.dataset.search.Search][] reference for more information on how to use it.
        """
        return Search[QueryType, JsonResultType, FieldsType, FormatType](
            query=query,
            format=format,
            dataset=cls.name(),
            fields=fields,
            size=size,
            include_isoform=include_isoform,
        )

    @overload
    @classmethod
    def fetch_one(
        cls, id: str, format: Literal["json"] = "json", parse: Literal[True] = True
    ) -> JsonResultType:
        ...

    @overload
    @classmethod
    def fetch_one(
        cls, id: str, format: FormatType, parse: Literal[False] = False
    ) -> IOBase:
        ...

    @classmethod
    def fetch_one(
        cls, id: str, format: str = "json", parse: bool = True
    ) -> JsonResultType | IOBase:
        """
        Fetches a single record from this dataset using its ID.

        Args:
            id : The ID of the record to fetch. The format of this will depend on the dataset.
            format : The format of the result. The available options will depend on the subclass you are using, but the type checker/autocomplete will enforce available options.
            parse : If true, parse the result into a JSON dictionary. Defaults to True.

        Returns:
            : If parse is True, a dictionary. Otherwise, a file object containing the results in the specified format.
        """
        res = requests.get(
            f"https://rest.uniprot.org/{cls.name()}/{id}.{format}", stream=True
        )
        res.raise_for_status()
        if parse and format == "json":
            return res.json()
        elif res.headers.get("Content-Encoding", None) == "gzip":
            return gzip.open(res.raw, "rb")
        else:
            return res.raw

    @classmethod
    def _type_args(cls) -> tuple[Type, ...]:
        """
        Returns the type arguments for this dataset.
        """
        return get_args(cls.__orig_bases__[0])  # type: ignore

    @classmethod
    def _allowed_query_fields(cls) -> set[str]:
        """
        Returns the type arguments for the queries to this dataset.
        """
        query_dict, _ = get_args(cls._query_type())
        return query_dict.__optional_keys__ | query_dict.__required_keys__ - {
            "and_",
            "or_",
            "not_",
        }

    @classmethod
    def _allowed_return_fields(cls) -> set[str]:
        """
        Returns the type arguments for the queries to this dataset.
        """
        return set(get_args(cls._field_type()))

    @classmethod
    def _query_type(cls):
        """
        Returns the type arguments for the queries to this dataset.
        """
        return cls._type_args()[0]

    @classmethod
    def _result_type(cls):
        """
        Returns the type of query results to this dataset.
        """
        return cls._type_args()[1]

    @classmethod
    def _field_type(cls):
        """
        Returns the type of allowed fields for search queries.
        """
        return cls._type_args()[2]

    @classmethod
    def _format_type(cls):
        """
        Returns the type of allowed formats for search queries.
        """
        return cls._type_args()[3]

    @classmethod
    def _allowed_formats(cls) -> set[str]:
        """
        Returns a set of allowed formats
        """
        return set(get_args(cls._format_type()))


class FetchManyClient(DatasetClient[QueryType, JsonResultType, FieldsType, FormatType]):
    """
    Dataset subclass for datasets that can be queried by multiple IDs. Not all datasets support this.
    """

    @classmethod
    def _bulk_id_param(cls) -> str:
        """
        The name of the GET query parameter used to define the list of IDs in a bulk query.
        """
        return cls._bulk_endpoint()

    @classmethod
    @abstractmethod
    def _bulk_endpoint(cls) -> str:
        """
        The name of the URL used to query a bulk list of IDs.
        """
        ...

    @overload
    @classmethod
    def fetch_many(
        cls,
        ids: Iterable[str],
        format: Literal["json"] = "json",
        parse: Literal[True] = True,
    ) -> Iterable[JsonResultType]:
        ...

    @overload
    @classmethod
    def fetch_many(
        cls, ids: Iterable[str], format: FormatType, parse: Literal[False] = False
    ) -> IOBase:
        ...

    @classmethod
    def fetch_many(
        cls,
        ids: Iterable[str],
        format: FormatType | Literal["json"] = "json",
        parse: bool = True,
    ) -> Iterable[JsonResultType] | IOBase:
        """
        Fetches multiple records using their accessions.

        Args:
            ids : The accessions to query
            format : The format to return the records. Defaults to "json".
            parse : Only supported for JSON. If True, parses the result instead of returning a raw file. Defaults to True.

        Returns:
            : If parse is True, a list of dictionaries. Otherwise, a file object containing the results in the specified format.
        """
        res = requests.get(
            f"https://rest.uniprot.org/{cls.name()}/{cls._bulk_endpoint()}",
            params={cls._bulk_id_param(): ",".join(ids), "format": format},
            stream=True,
        )
        res.raise_for_status()
        if parse and format == "json":
            return res.json()["results"]
        elif res.headers.get("Content-Encoding", None) == "gzip":
            return gzip.open(res.raw, "rb")
        else:
            return res.raw



from typing import Any, Mapping

from typing_extensions import Literal

from unipressed.dataset.core import DatasetClient
from unipressed.dataset.generated_types.keywords import KeywordsFields, KeywordsQuery

KeywordsFormats = Literal["json", "tsv", "xlsx", "list", "obo"]


class KeywordsClient(
    DatasetClient[
        KeywordsQuery,
        Mapping[str, Any],
        KeywordsFields,
        KeywordsFormats,
    ]
):
    @classmethod
    def name(cls):
        return "keywords"

    @classmethod
    def _id_field(cls, record: Mapping[str, Any]) -> str:
        return record["keyword"]["id"]



from typing import Any, Mapping

from typing_extensions import Literal

from unipressed.dataset.core import FetchManyClient
from unipressed.dataset.generated_types.taxonomy import TaxonomyFields, TaxonomyQuery

TaxonomyFormats = Literal["json", "tsv", "xlsx", "list"]


class TaxonomyClient(
    FetchManyClient[
        TaxonomyQuery,
        Mapping[str, Any],
        TaxonomyFields,
        TaxonomyFormats,
    ]
):
    @classmethod
    def name(cls):
        return "taxonomy"

    @classmethod
    def _bulk_endpoint(cls):
        return "taxonIds"

    @classmethod
    def _id_field(cls, record: Mapping[str, Any]) -> str:
        return record["taxonId"]



from __future__ import annotations

from datetime import date
from typing import Iterable, Union

from typing_extensions import Literal, NotRequired, TypeAlias, TypedDict

Rank: TypeAlias = Literal[
    "SUPERKINGDOM",
    "KINGDOM",
    "SUBKINGDOM",
    "SUPERPHYLUM",
    "PHYLUM",
    "SUBPHYLUM",
    "SUPERCLASS",
    "CLASS",
    "SUBCLASS",
    "INFRACLASS",
    "COHORT",
    "SUBCOHORT",
    "SUPERORDER",
    "ORDER",
    "SUBORDER",
    "INFRAORDER",
    "PARVORDER",
    "SUPERFAMILY",
    "FAMILY",
    "SUBFAMILY",
    "TRIBE",
    "SUBTRIBE",
    "GENUS",
    "SUBGENUS",
    "SPECIES_GROUP",
    "SPECIES_SUBGROUP",
    "SPECIES",
    "SUBSPECIES",
    "VARIETAS",
    "FORMA",
    "NO_RANK",
]


class TaxonomyQueryDict(TypedDict):
    and_: NotRequired[Iterable["TaxonomyQuery"]]
    "Two or more filters that must both be satisfied"
    or_: NotRequired[Iterable["TaxonomyQuery"]]
    "Two or more filters, any of which can be satisfied"
    not_: NotRequired[Iterable["TaxonomyQuery"]]
    "Negate a filter"
    tax_id: NotRequired[int]
    "Taxon ID\ne.g. 85621"
    scientific: NotRequired[str]
    "Scientific name\ne.g. 16SrII"
    common: NotRequired[str]
    "Common name\ne.g. sample name"
    mnemonic: NotRequired[str]
    "Mnemonic\ne.g. sample mnemonic"
    rank: NotRequired[Rank]
    "Rank\ne.g. SPECIES_GROUP\n* SUPERKINGDOM: Superkingdom\n* KINGDOM: Kingdom\n* SUBKINGDOM: Subkingdom\n* SUPERPHYLUM: Superphylum\n* PHYLUM: Phylum\n* SUBPHYLUM: Subphylum\n* SUPERCLASS: Superclass\n* CLASS: Class\n* SUBCLASS: Subclass\n* INFRACLASS: Infraclass\n* COHORT: Cohort\n* SUBCOHORT: Subcohort\n* SUPERORDER: Superorder\n* ORDER: Order\n* SUBORDER: Suborder\n* INFRAORDER: Infraorder\n* PARVORDER: Parvorder\n* SUPERFAMILY: Superfamily\n* FAMILY: Family\n* SUBFAMILY: Subfamily\n* TRIBE: Tribe\n* SUBTRIBE: Subtribe\n* GENUS: Genus\n* SUBGENUS: Subgenus\n* SPECIES_GROUP: Species group\n* SPECIES_SUBGROUP: Species subgroup\n* SPECIES: Species\n* SUBSPECIES: Subspecies\n* VARIETAS: Varietas\n* FORMA: Forma\n* NO_RANK: No rank"
    strain: NotRequired[str]
    "Strain\ne.g. SPECIES_GROUP"
    host: NotRequired[int]
    "Virus host\ne.g. 85621"
    linked: NotRequired[bool]
    "With external info\ne.g. true"
    parent: NotRequired[str]
    "Parent\ne.g. 9606"
    ancestor: NotRequired[str]
    "Ancestor\ne.g. 85621"


TaxonomyQuery: TypeAlias = Union[TaxonomyQueryDict, str]
TaxonomyTaxonomy: TypeAlias = Literal[
    "id",
    "mnemonic",
    "scientific_name",
    "common_name",
    "synonyms",
    "other_names",
    "rank",
    "reviewed",
    "lineage",
    "strains",
    "parent",
    "hosts",
    "links",
    "statistics",
]
TaxonomyFields: TypeAlias = Literal[TaxonomyTaxonomy,]



from typing import Any, Mapping

from typing_extensions import Literal

from unipressed.dataset.core import FetchManyClient
from unipressed.dataset.generated_types.uniref import UnirefFields, UnirefQuery

UnirefFormats = Literal["fasta", "tsv", "json", "xlsx", "list"]


class UnirefClient(
    FetchManyClient[UnirefQuery, Mapping[str, Any], UnirefFields, UnirefFormats]
):
    @classmethod
    def name(cls):
        return "uniref"

    @classmethod
    def _bulk_endpoint(cls):
        return "ids"



from __future__ import annotations

import dataclasses
import datetime
import gzip
from collections.abc import Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Generic, Iterable, TextIO

import requests
from typing_extensions import Literal, TypedDict

from unipressed.dataset.type_vars import (
    FieldsType,
    FormatType,
    JsonResultType,
    QueryType,
)

if TYPE_CHECKING:
    from xml.etree.ElementTree import Element


@dataclasses.dataclass
class SearchRequest:
    """
    Low level options for a Uniprot request, which the user generally doesn't need to consider
    """

    high_level: Search
    compressed: bool = True
    cursor: str | None = None

    def params(self) -> dict[str, str]:
        """
        :return:
        """
        params = dataclasses.asdict(self)
        params.pop("high_level")
        return {**params, **self.high_level._params()}

    def to_request(self) -> requests.PreparedRequest:
        return requests.Request(
            method="GET",
            url=f"https://rest.uniprot.org/{self.high_level.dataset}/search",
            params=self.params(),
        ).prepare()


def serialize_query(query: Any, level: int = 0) -> str:
    """
    Recursively converts a query object into a Uniprot query string

    Args:
        query (Any): A dict, str, or any other type to convert to string
        level (int, optional): The current depth into the query
    """
    if isinstance(query, dict):
        if len(query) > 1:
            # If we ever find a dictionary with more than one key, we treat it as an AND
            return serialize_query(
                {"and_": [{key: value} for key, value in query.items()]}
            )
        key, value = next(iter(query.items()))
        if key == "and_":
            ret = " AND ".join([serialize_query(it, level=level + 1) for it in value])
        elif key == "or_":
            ret = " OR ".join([serialize_query(it, level=level + 1) for it in value])
        elif key == "not_":
            ret = f"NOT {serialize_query(value, level=level+1)}"
        else:
            return f"{key}:{serialize_query(value)}"

        if level > 0:
            return f"({ret})"
        else:
            return ret
    elif isinstance(query, str):
        # Strings can be used directly as queries
        return query
    elif isinstance(query, Sequence) and len(query) == 2:
        # A length 2 sequence is treated as a range
        a, b = query
        return f"[{serialize_query(a)} TO {serialize_query(b)}]"
    elif isinstance(query, datetime.date):
        return query.strftime("%Y-%m-%d")
    elif isinstance(query, bool):
        return str(query).lower()
    else:
        return str(query)


class JsonRecord(TypedDict, total=False):
    primaryAccession: str


@dataclass
class Search(
    Generic[QueryType, JsonResultType, FieldsType, FormatType],
):
    """
    Class coordinating search requests. Generally you shouldn't instantiate this manually, but should rather use [unipressed.DatasetClient.search][] to do so.
    """

    query: QueryType
    "A query that filters the returned proteins. Either a string if you have an existing query and want to bypass the type checker, or a dict if you want code completion and validation."
    dataset: str = "uniprotkb"
    """
    The Uniprot database. You are advised to use the appropriate subclass (e.g. [unipressed.UniprotkbSearch][] where available.
    """
    format: FormatType | Literal["json"] = "json"
    "Format for the returned data. Defaults to `'json'`."
    fields: Iterable[FieldsType] | None = None
    "An iterable of fields to return in the result object."
    include_isoform: bool = True
    "For uniprotkb only: if `True`, returns isoforms other than just the canonical isoform."
    size: int = 500
    "The size of each page. Changing this may slightly affect performance, and you aren't recommended to do so unless you have a reason to."

    def _params(self) -> dict[str, str]:
        """
        Returns the URL query parameters that can be derived from this object
        """
        ret = {
            "query": serialize_query(self.query),
            "dataset": self.dataset,
            "format": self.format,
            "includeIsoform": str(self.include_isoform).lower(),
            "size": str(self.size),
        }
        if self.fields:
            ret["fields"] = ",".join(self.fields)
        return ret

    def each_response(self) -> Iterable[requests.Response]:
        """
        Returns a generator of [`requests.Response`](https://requests.readthedocs.io/en/latest/api/#requests.Response) objects, one for each page of the result.
        """
        session = requests.Session()
        request = SearchRequest(self).to_request()
        while True:
            response = session.send(request, stream=True)
            yield response
            link: dict[str, Any] | None = response.links.get("next")
            if link is not None:
                # pyright: ignore
                request = requests.Request("GET", link["url"]).prepare()
            else:
                break

    def each_page(self) -> Iterable[TextIO]:
        """
        Returns a generator of unzipped [file objects](https://docs.python.org/3/library/io.html#io.TextIOBase), one for each page of the result.
        """
        for response in self.each_response():
            response.raise_for_status()
            yield gzip.open(response.raw, mode="rt", encoding=response.encoding)

    def each_record(self) -> Iterable[Any]:
        """
        Returns a generator of records, which are defined by the format field of the original request.
        For example, with `format="json"`, this will be an iterator over dictionaries parsed from JSON.
        """
        parser = getattr(self, f"_each_{self.format}", None)
        if parser is not None:
            for page in self.each_page():
                yield from parser(page)
        else:
            raise NotImplementedError(
                f'No parser is implemented for the "{self.format}" format.'
            )

    @staticmethod
    def _each_json(page: TextIO) -> Iterable[JsonRecord]:
        import json

        parsed = json.load(page)
        yield from parsed["results"]

    @staticmethod
    def _each_tsv(page: TextIO) -> Iterable[dict[str, str]]:
        import csv

        reader = csv.DictReader(page, delimiter="\t")
        yield from reader

    @staticmethod
    def _each_xml(page: TextIO) -> Iterable[Element]:
        from xml.etree import ElementTree

        tree = ElementTree.parse(page)
        yield from tree.getroot().findall("{http://uniprot.org/uniprot}entry")

    @staticmethod
    def _each_list(page: TextIO) -> Iterable[str]:
        yield from page.read().splitlines()



from __future__ import annotations

from datetime import date
from typing import Iterable, Union

from typing_extensions import Literal, NotRequired, TypeAlias, TypedDict

Category: TypeAlias = Literal[
    "technical_term",
    "ptm",
    "molecular_function",
    "ligand",
    "domain",
    "disease",
    "developmental_stage",
    "coding_sequence_diversity",
    "cellular_component",
    "biological_process",
]


class KeywordsQueryDict(TypedDict):
    and_: NotRequired[Iterable["KeywordsQuery"]]
    "Two or more filters that must both be satisfied"
    or_: NotRequired[Iterable["KeywordsQuery"]]
    "Two or more filters, any of which can be satisfied"
    not_: NotRequired[Iterable["KeywordsQuery"]]
    "Negate a filter"
    name: NotRequired[str]
    "Name\ne.g. 2Fe-2S"
    keyword_id: NotRequired[str]
    "Keyword [AC]\ne.g. KW-0001"
    category: NotRequired[Category]
    "Category\ne.g. Domain\n* technical_term: Technical term [KW-9990]\n* ptm: PTM [KW-9991]\n* molecular_function: Molecular function [KW-9992]\n* ligand: Ligand [KW-9993]\n* domain: Domain [KW-9994]\n* disease: Disease [KW-9995]\n* developmental_stage: Developmental stage [KW-9996]\n* coding_sequence_diversity: Coding sequence diversity [KW-9997]\n* cellular_component: Cellular component [KW-9998]\n* biological_process: Biological process [KW-9999]"


KeywordsQuery: TypeAlias = Union[KeywordsQueryDict, str]
KeywordsKeyword: TypeAlias = Literal[
    "id",
    "name",
    "definition",
    "category",
    "synonyms",
    "gene_ontologies",
    "links",
    "children",
    "parents",
    "statistics",
]
KeywordsFields: TypeAlias = Literal[KeywordsKeyword,]



from __future__ import annotations

from datetime import date
from typing import Iterable, Union

from typing_extensions import Literal, NotRequired, TypeAlias, TypedDict


class DatabaseQueryDict(TypedDict):
    and_: NotRequired[Iterable["DatabaseQuery"]]
    "Two or more filters that must both be satisfied"
    or_: NotRequired[Iterable["DatabaseQuery"]]
    "Two or more filters, any of which can be satisfied"
    not_: NotRequired[Iterable["DatabaseQuery"]]
    "Negate a filter"
    name: NotRequired[str]
    "Name\ne.g. Arabidopsis Information Portal"
    id: NotRequired[str]
    "Database [AC]\ne.g. DB-0236"


DatabaseQuery: TypeAlias = Union[DatabaseQueryDict, str]
DatabaseCrossReference: TypeAlias = Literal[
    "id",
    "name",
    "abbrev",
    "pubmed_id",
    "doi_id",
    "link_type",
    "servers",
    "dbUrl",
    "category",
    "statistics",
]
DatabaseFields: TypeAlias = Literal[DatabaseCrossReference,]



from __future__ import annotations

from datetime import date
from typing import Iterable, Union

from typing_extensions import Literal, NotRequired, TypeAlias, TypedDict

Identity: TypeAlias = Literal["1.0", "0.9", "0.5"]


class UnirefQueryDict(TypedDict):
    and_: NotRequired[Iterable["UnirefQuery"]]
    "Two or more filters that must both be satisfied"
    or_: NotRequired[Iterable["UnirefQuery"]]
    "Two or more filters, any of which can be satisfied"
    not_: NotRequired[Iterable["UnirefQuery"]]
    "Negate a filter"
    id: NotRequired[str]
    "UniRef ID\ne.g. UniRef100_A0A001"
    name: NotRequired[str]
    "Cluster name\ne.g. sample name"
    identity: NotRequired[Identity]
    "Sequence identity\ne.g. sample identity\n* 1.0: 100%\n* 0.9: 90%\n* 0.5: 50%"
    count: NotRequired[
        tuple[
            Union[
                int,
                Literal["*",],
            ],
            Union[
                int,
                Literal["*",],
            ],
        ]
    ]
    "Cluster size\ne.g. [100 TO 300]"
    length: NotRequired[
        tuple[
            Union[
                int,
                Literal["*",],
            ],
            Union[
                int,
                Literal["*",],
            ],
        ]
    ]
    "Sequence length\ne.g. [100 TO 300]"
    date_modified: NotRequired[
        tuple[
            Union[
                date,
                Literal["*",],
            ],
            Union[
                date,
                Literal["*",],
            ],
        ]
    ]
    "Date of last modification\ne.g. [2011-10-10 TO 2019-10-10]"
    uniprot_id: NotRequired[str]
    "UniProtKB ID/AC\ne.g. sample uniprot id"
    upi: NotRequired[str]
    "UniParc ID\ne.g. UPI0123456789"
    taxonomy_name: NotRequired[str]
    "Taxonomy [OC]\ne.g. sample name"
    taxonomy_id: NotRequired[str]
    "Taxonomy id"
    cluster: NotRequired[str]
    "Related clusters\ne.g. UniRef100_A0A001"


UnirefQuery: TypeAlias = Union[UnirefQueryDict, str]
UnirefNamesTaxonomy: TypeAlias = Literal[
    "id", "name", "common_taxon", "common_taxonid", "organism_id", "organism"
]
UnirefSequences: TypeAlias = Literal["identity", "length", "sequence"]
UnirefMiscellaneous: TypeAlias = Literal["types", "members", "count"]
UnirefDateOf: TypeAlias = Literal["date_modified",]
UnirefFields: TypeAlias = Literal[
    UnirefNamesTaxonomy, UnirefSequences, UnirefMiscellaneous, UnirefDateOf
]



from typing import Any, Mapping

from typing_extensions import Literal

from unipressed.dataset.core import FetchManyClient
from unipressed.dataset.generated_types.uniparc import UniparcFields, UniparcQuery

UniparcFormats = Literal["fasta", "tsv", "xlsx", "json", "xml"]


class UniparcClient(
    FetchManyClient[UniparcQuery, Mapping[str, Any], UniparcFields, UniparcFormats]
):
    @classmethod
    def name(cls):
        return "uniparc"

    @classmethod
    def _bulk_endpoint(cls):
        return "upis"

    @classmethod
    def _id_field(cls, record: Mapping[str, Any]) -> str:
        return record["uniParcId"]



from typing import Any, Mapping

from typing_extensions import Literal

from unipressed.dataset.core import DatasetClient
from unipressed.dataset.generated_types.database import DatabaseFields, DatabaseQuery

DatabaseFormats = Literal["json"]


class DatabaseClient(
    DatasetClient[
        DatabaseQuery,
        Mapping[str, Any],
        DatabaseFields,
        DatabaseFormats,
    ]
):
    @classmethod
    def name(cls):
        return "database"



from __future__ import annotations

from typing import Type

from unipressed.dataset.arba import ArbaClient
from unipressed.dataset.citations import CitationsClient
from unipressed.dataset.core import DatasetClient
from unipressed.dataset.database import DatabaseClient
from unipressed.dataset.diseases import DiseasesClient
from unipressed.dataset.keywords import KeywordsClient
from unipressed.dataset.locations import LocationsClient
from unipressed.dataset.proteomes import ProteomesClient
from unipressed.dataset.taxonomy import TaxonomyClient
from unipressed.dataset.uniparc import UniparcClient
from unipressed.dataset.uniprotkb import UniprotkbClient
from unipressed.dataset.uniref import UnirefClient
from unipressed.dataset.unirule import UniruleClient

all_clients: list[Type[DatasetClient]] = [
    ArbaClient,
    CitationsClient,
    DatabaseClient,
    DiseasesClient,
    KeywordsClient,
    LocationsClient,
    ProteomesClient,
    TaxonomyClient,
    UniparcClient,
    UniprotkbClient,
    UnirefClient,
    UniruleClient,
]



from __future__ import annotations

from datetime import date
from typing import Iterable, Union

from typing_extensions import Literal, NotRequired, TypeAlias, TypedDict


class ArbaQueryDict(TypedDict):
    and_: NotRequired[Iterable["ArbaQuery"]]
    "Two or more filters that must both be satisfied"
    or_: NotRequired[Iterable["ArbaQuery"]]
    "Two or more filters, any of which can be satisfied"
    not_: NotRequired[Iterable["ArbaQuery"]]
    "Negate a filter"
    protein_name: NotRequired[str]
    "Protein Name [DE]\ne.g. mas5"
    organism: NotRequired[str]
    "Organism [OS]\ne.g. saccharomyces"
    taxonomy: NotRequired[str]
    "Taxonomy [OC]\ne.g. human"
    ec: NotRequired[str]
    "Enzyme classification [EC]\ne.g. 1.1.2.3"
    cc_cofactor: NotRequired[str]
    "Cofactor\ne.g. 29105"
    cc_catalytic_activity: NotRequired[str]
    "Catalytic activity\ne.g. tyrosine"
    cc_activity_regulation: NotRequired[str]
    "Activity regulation\ne.g. inhibited"
    cc_pathway: NotRequired[str]
    "Pathway\ne.g. metabolism"
    cc_subcellular_location: NotRequired[str]
    "Subcellular location term\ne.g. membrane"
    cc_subcellular_location_note: NotRequired[str]
    "Note\ne.g. membrane"
    cc_domain: NotRequired[str]
    "Domain comments [CC]\ne.g. conformation"
    family: NotRequired[str]
    "Protein family\ne.g. pa28"
    cc_similarity: NotRequired[str]
    "Comment similarity\ne.g. phosphatase"
    keyword: NotRequired[str]
    "Keyword [KW]\ne.g. chromosomal"


ArbaQuery: TypeAlias = Union[ArbaQueryDict, str]
ArbaArba: TypeAlias = Literal[
    "rule_id", "statistics", "taxonomic_scope", "annotation_covered"
]
ArbaFields: TypeAlias = Literal[ArbaArba,]



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Callable, Iterable

import requests
from typing_extensions import Literal, ParamSpec, TypeAlias, TypedDict, TypeVar

import unipressed.id_mapping.types as id_types
from unipressed.util import iter_pages

UniprotStatus: TypeAlias = Literal[
    # See https://github.com/ebi-uniprot/uniprot-rest-api/blob/master/idmapping-rest/src/main/java/org/uniprot/api/idmapping/controller/response/JobStatus.java
    "NEW",
    "RUNNING",
    "FINISHED",
    "ERROR",
]

Param1 = ParamSpec("Param1")
Param2 = ParamSpec("Param2")
Ret1 = TypeVar("Ret1")
Ret2 = TypeVar("Ret2")


def copy_signature(
    f: Callable[Param1, Ret1]
) -> Callable[[Callable[Param2, Ret2]], Callable[Param1, Ret2]]:
    """
    Copies the argument signature from function f and applies it to the decorated function, but keeps the return value
    """

    def _inner(f: Callable[Param2, Ret2]):
        return f

    return _inner  # type: ignore


class IdMappingError(Exception):
    pass


@dataclass
class IdMappingClient:
    """
    Client for submitting requests to convert between identifiers.
    """

    @classmethod
    def _submit(
        cls, source: str, dest: str, ids: Iterable[str], taxon_id: int | None
    ) -> requests.Response:
        data: dict[str, Any] = {"ids": ",".join(ids), "from": source, "to": dest}
        if taxon_id is not None:
            data["taxId"] = taxon_id
        return requests.post(
            "https://rest.uniprot.org/idmapping/run",
            data=data,
        )

    @copy_signature(id_types.SubmitDummyClass.submit)
    @classmethod
    def submit(
        cls, source: str, dest: str, ids: Iterable[str], taxon_id: int | None = None
    ) -> IdMappingJob:
        """
        Submits this ID mapping request to the UniProt server, and returns a new object that can be used to access the results

        Args:
            source: Name of the database from which the ids originate
            dest: Name of the database that the IDs will be converted to
            ids: Set of IDs to convert
            taxon_id: Optionally, a UniProt taxon ID to restrict the results to. For example, you can use `4932` to
                restrict the results to Baker's yest (https://www.uniprot.org/taxonomy/4932).
        """
        res = cls._submit(source, dest, ids, taxon_id=taxon_id)
        job_id = res.json()["jobId"]
        return IdMappingJob(job_id)


@dataclass
class IdMappingJob:
    """
    Object that tracks the status and results of an ID mapping request that has been sent to the UniProt server.
    """

    job_id: str

    def get_status(self) -> UniprotStatus:
        """
        Returns a string describing the status of the job (ie if it has completed or not).
        """
        data = requests.get(
            f"https://rest.uniprot.org/idmapping/status/{self.job_id}"
        ).json()
        if "jobStatus" in data:
            # This is the correct behaviour
            return data["jobStatus"]
        elif "results" in data:
            # Sometimes UniProt returns the actual results even though we asked for the status, see https://github.com/multimeric/Unipressed/issues/21
            return "FINISHED"
        else:
            raise IdMappingError(
                f"Unknown response returned by UniProt. Received {data}"
            )

    def each_result(self) -> Iterable[IdMappingResult]:
        """
        Returns a generator over dictionaries of results, one for each input ID.
        """
        res = requests.get(f"https://rest.uniprot.org/idmapping/results/{self.job_id}")
        for page in iter_pages(res):
            parsed = page.json()
            if "results" not in parsed:
                raise IdMappingError(
                    "UniProt has not yet processed the results, consider using time.sleep() to wait until they are complete."
                )
            else:
                yield from parsed["results"]


IdMappingResult = TypedDict("IdMappingResult", {"from": str, "to": str})



from typing import Any, Mapping

from typing_extensions import Literal

from unipressed.dataset.core import DatasetClient
from unipressed.dataset.generated_types.arba import ArbaFields, ArbaQuery

ArbaFormats = Literal["json", "list"]


class ArbaClient(DatasetClient[ArbaQuery, Mapping[str, Any], ArbaFields, ArbaFormats]):
    @classmethod
    def name(cls):
        return "arba"

    @classmethod
    def _id_field(cls, record: Any) -> str:
        return record["uniRuleId"]



from __future__ import annotations

from datetime import date
from typing import Iterable, Union

from typing_extensions import Literal, NotRequired, TypeAlias, TypedDict


class LocationsQueryDict(TypedDict):
    and_: NotRequired[Iterable["LocationsQuery"]]
    "Two or more filters that must both be satisfied"
    or_: NotRequired[Iterable["LocationsQuery"]]
    "Two or more filters, any of which can be satisfied"
    not_: NotRequired[Iterable["LocationsQuery"]]
    "Negate a filter"
    name: NotRequired[str]
    "Name\ne.g. A band"
    id: NotRequired[str]
    "Location [AC]\ne.g. SL-0476"


LocationsQuery: TypeAlias = Union[LocationsQueryDict, str]
LocationsSubcellularLocation: TypeAlias = Literal[
    "id",
    "name",
    "definition",
    "category",
    "keyword",
    "synonyms",
    "content",
    "gene_ontologies",
    "note",
    "references",
    "links",
    "is_a",
    "part_of",
    "statistics",
]
LocationsFields: TypeAlias = Literal[LocationsSubcellularLocation,]



from __future__ import annotations

from datetime import date
from typing import Iterable, Union

from typing_extensions import Literal, NotRequired, TypeAlias, TypedDict


class CitationsQueryDict(TypedDict):
    and_: NotRequired[Iterable["CitationsQuery"]]
    "Two or more filters that must both be satisfied"
    or_: NotRequired[Iterable["CitationsQuery"]]
    "Two or more filters, any of which can be satisfied"
    not_: NotRequired[Iterable["CitationsQuery"]]
    "Negate a filter"
    title: NotRequired[str]
    "Title\ne.g. sample title"
    author: NotRequired[str]
    "Author\ne.g. sample author name"
    journal: NotRequired[str]
    "Journal\ne.g. sample journal"
    published: NotRequired[str]
    "Year published\ne.g. sample"
    id: NotRequired[str]
    "Citation Id\ne.g. 15772651"
    doi: NotRequired[str]
    "DOI\ne.g. 10.1007/s11882-009-0055-9"


CitationsQuery: TypeAlias = Union[CitationsQueryDict, str]
CitationsLiterature: TypeAlias = Literal[
    "id",
    "doi",
    "title",
    "authors",
    "authoring_group",
    "journal",
    "publication_date",
    "first_page",
    "last_page",
    "volume",
    "reference",
    "lit_abstract",
    "statistics",
]
CitationsFields: TypeAlias = Literal[CitationsLiterature,]



from typing import Any, Mapping

from typing_extensions import Literal

from unipressed.dataset.core import DatasetClient
from unipressed.dataset.generated_types.proteomes import ProteomesFields, ProteomesQuery

ProteomesFormats = Literal["json", "tsv", "xlsx", "list", "xml"]


class ProteomesClient(
    DatasetClient[
        ProteomesQuery,
        Mapping[str, Any],
        ProteomesFields,
        ProteomesFormats,
    ]
):
    @classmethod
    def name(cls):
        return "proteomes"



from __future__ import annotations

from datetime import date
from typing import Iterable, Union

from typing_extensions import Literal, NotRequired, TypeAlias, TypedDict

Existence: TypeAlias = Literal["1", "2", "3", "4", "5"]
Organelle: TypeAlias = Literal[
    "mitochondrion",
    "plasmid",
    "plastid",
    "chloroplast",
    "cyanelle",
    "apicoplast",
    "organellar chromatophore",
    "non-photosynthetic plastid",
    "nucleomorph",
    "hydrogenosome",
]


class UniprotkbQueryDict(TypedDict):
    and_: NotRequired[Iterable["UniprotkbQuery"]]
    "Two or more filters that must both be satisfied"
    or_: NotRequired[Iterable["UniprotkbQuery"]]
    "Two or more filters, any of which can be satisfied"
    not_: NotRequired[Iterable["UniprotkbQuery"]]
    "Negate a filter"
    accession: NotRequired[str]
    "UniProtKB AC\ne.g. P12345"
    id: NotRequired[str]
    "Entry Name [ID]\ne.g. P53_HUMAN"
    protein_name: NotRequired[str]
    "Protein Name [DE]\ne.g. mas5"
    gene: NotRequired[str]
    "Gene Name [GN]\ne.g. ydj1"
    organism_name: NotRequired[str]
    "Organism [OS]\ne.g. saccharomyces"
    organism_id: NotRequired[str]
    "Organism id"
    taxonomy_name: NotRequired[str]
    "Taxonomy [OC]\ne.g. human"
    taxonomy_id: NotRequired[str]
    "Taxonomy id"
    virus_host_name: NotRequired[str]
    "Virus host\ne.g. human"
    virus_host_id: NotRequired[str]
    "Virus host id"
    existence: NotRequired[Existence]
    "Protein Existence [PE]\ne.g. 1\n* 1: Evidence at protein level\n* 2: Evidence at transcript level\n* 3: Inferred from homology\n* 4: Predicted\n* 5: Uncertain"
    ec: NotRequired[str]
    "Enzyme classification [EC]\ne.g. 1.1.2.3"
    cc_cofactor_chebi: NotRequired[str]
    "Cc cofactor chebi\ne.g. 29105"
    cc_cofactor_chebi_exp: NotRequired[str]
    "Cc cofactor chebi exp\ne.g. 29105"
    cc_cofactor_note: NotRequired[str]
    "Cc cofactor note\ne.g. subunit"
    cc_cofactor_note_exp: NotRequired[str]
    "Cc cofactor note exp\ne.g. subunit"
    cc_bpcp: NotRequired[str]
    "Cc bpcp\ne.g. prosthetic"
    cc_bpcp_exp: NotRequired[str]
    "Cc bpcp exp\ne.g. prosthetic"
    cc_bpcp_absorption: NotRequired[str]
    "Cc bpcp absorption\ne.g. prosthetic"
    cc_bpcp_absorption_exp: NotRequired[str]
    "Cc bpcp absorption exp\ne.g. prosthetic"
    cc_bpcp_kinetics: NotRequired[str]
    "Cc bpcp kinetics\ne.g. aspartate"
    cc_bpcp_kinetics_exp: NotRequired[str]
    "Cc bpcp kinetics exp\ne.g. aspartate"
    cc_bpcp_ph_dependence: NotRequired[str]
    "Cc bpcp ph dependence\ne.g. optimum"
    cc_bpcp_ph_dependence_exp: NotRequired[str]
    "Cc bpcp ph dependence exp\ne.g. optimum"
    cc_bpcp_redox_potential: NotRequired[str]
    "Cc bpcp redox potential\ne.g. siroheme"
    cc_bpcp_redox_potential_exp: NotRequired[str]
    "Cc bpcp redox potential exp\ne.g. siroheme"
    cc_bpcp_temp_dependence: NotRequired[str]
    "Cc bpcp temp dependence\ne.g. *"
    cc_bpcp_temp_dependence_exp: NotRequired[str]
    "Cc bpcp temp dependence exp\ne.g. *"
    cc_catalytic_activity: NotRequired[str]
    "Cc catalytic activity\ne.g. tyrosine"
    cc_catalytic_activity_exp: NotRequired[str]
    "Cc catalytic activity exp\ne.g. tyrosine"
    cc_activity_regulation: NotRequired[str]
    "Cc activity regulation\ne.g. inhibited"
    cc_activity_regulation_exp: NotRequired[str]
    "Cc activity regulation exp\ne.g. inhibited"
    cc_function: NotRequired[str]
    "Cc function\ne.g. enzyme"
    cc_function_exp: NotRequired[str]
    "Cc function exp\ne.g. enzyme"
    cc_caution: NotRequired[str]
    "Cc caution\ne.g. kinase"
    cc_caution_exp: NotRequired[str]
    "Cc caution exp\ne.g. kinase"
    ft_sites: NotRequired[str]
    "Ft sites\ne.g. translocation"
    ft_sites_exp: NotRequired[str]
    "Ft sites exp\ne.g. translocation"
    ft_act_site: NotRequired[str]
    "Ft act site\ne.g. phosphocysteine"
    ft_act_site_exp: NotRequired[str]
    "Ft act site exp\ne.g. phosphocysteine"
    ft_binding: NotRequired[str]
    "Ft binding\ne.g. phosphocysteine"
    ft_binding_exp: NotRequired[str]
    "Ft binding exp\ne.g. phosphocysteine"
    ft_site: NotRequired[str]
    "Ft site\ne.g. phosphocysteine"
    ft_site_exp: NotRequired[str]
    "Ft site exp\ne.g. phosphocysteine"
    ft_dna_bind: NotRequired[str]
    "Ft dna bind\ne.g. *"
    ft_dna_bind_exp: NotRequired[str]
    "Ft dna bind exp\ne.g. *"
    cc_pathway: NotRequired[str]
    "Cc pathway\ne.g. metabolism"
    cc_pathway_exp: NotRequired[str]
    "Cc pathway exp\ne.g. metabolism"
    cc_miscellaneous: NotRequired[str]
    "Cc miscellaneous\ne.g. abscisic"
    cc_miscellaneous_exp: NotRequired[str]
    "Cc miscellaneous exp\ne.g. abscisic"
    cc_scl_term: NotRequired[str]
    "Cc scl term\ne.g. membrane"
    cc_scl_term_exp: NotRequired[str]
    "Cc scl term exp\ne.g. membrane"
    cc_scl_note: NotRequired[str]
    "Cc scl note\ne.g. membrane"
    cc_scl_note_exp: NotRequired[str]
    "Cc scl note exp\ne.g. membrane"
    ft_transmem: NotRequired[str]
    "Ft transmem\ne.g. forming"
    ft_transmem_exp: NotRequired[str]
    "Ft transmem exp\ne.g. forming"
    ft_topo_dom: NotRequired[str]
    "Ft topo dom\ne.g. forming"
    ft_topo_dom_exp: NotRequired[str]
    "Ft topo dom exp\ne.g. forming"
    ft_intramem: NotRequired[str]
    "Ft intramem\ne.g. forming"
    ft_intramem_exp: NotRequired[str]
    "Ft intramem exp\ne.g. forming"
    cc_disease: NotRequired[str]
    "Cc disease\ne.g. nephrotic"
    cc_disease_exp: NotRequired[str]
    "Cc disease exp\ne.g. nephrotic"
    cc_allergen: NotRequired[str]
    "Cc allergen\ne.g. allergic"
    cc_allergen_exp: NotRequired[str]
    "Cc allergen exp\ne.g. allergic"
    cc_toxic_dose: NotRequired[str]
    "Cc toxic dose\ne.g. intracistenal"
    cc_toxic_dose_exp: NotRequired[str]
    "Cc toxic dose exp\ne.g. intracistenal"
    cc_biotechnology: NotRequired[str]
    "Cc biotechnology\ne.g. vaccine"
    cc_biotechnology_exp: NotRequired[str]
    "Cc biotechnology exp\ne.g. vaccine"
    cc_pharmaceutical: NotRequired[str]
    "Cc pharmaceutical\ne.g. peptide"
    cc_pharmaceutical_exp: NotRequired[str]
    "Cc pharmaceutical exp\ne.g. peptide"
    cc_disruption_phenotype: NotRequired[str]
    "Cc disruption phenotype\ne.g. infected"
    cc_disruption_phenotype_exp: NotRequired[str]
    "Cc disruption phenotype exp\ne.g. infected"
    ft_mutagen: NotRequired[str]
    "Ft mutagen\ne.g. phosphatase"
    ft_mutagen_exp: NotRequired[str]
    "Ft mutagen exp\ne.g. phosphatase"
    cc_ptm: NotRequired[str]
    "Cc ptm\ne.g. mitosis"
    cc_ptm_exp: NotRequired[str]
    "Cc ptm exp\ne.g. mitosis"
    ft_mod_res: NotRequired[str]
    "Ft mod res\ne.g. phosphoserine"
    ft_mod_res_exp: NotRequired[str]
    "Ft mod res exp\ne.g. phosphoserine"
    ft_lipid: NotRequired[str]
    "Ft lipid\ne.g. cysteine"
    ft_lipid_exp: NotRequired[str]
    "Ft lipid exp\ne.g. cysteine"
    ft_carbohyd: NotRequired[str]
    "Ft carbohyd\ne.g. cysteine"
    ft_carbohyd_exp: NotRequired[str]
    "Ft carbohyd exp\ne.g. cysteine"
    ft_disulfid: NotRequired[str]
    "Ft disulfid\ne.g. reversible"
    ft_disulfid_exp: NotRequired[str]
    "Ft disulfid exp\ne.g. reversible"
    ft_crosslnk: NotRequired[str]
    "Ft crosslnk\ne.g. lysine"
    ft_crosslnk_exp: NotRequired[str]
    "Ft crosslnk exp\ne.g. lysine"
    ft_molecule_processing: NotRequired[str]
    "Ft molecule processing\ne.g. disulfide"
    ft_molecule_processing_exp: NotRequired[str]
    "Ft molecule processing exp\ne.g. disulfide"
    ft_chain: NotRequired[str]
    "Ft chain\ne.g. kinase"
    ft_chain_exp: NotRequired[str]
    "Ft chain exp\ne.g. kinase"
    ft_init_met: NotRequired[str]
    "Ft init met\ne.g. Removed"
    ft_init_met_exp: NotRequired[str]
    "Ft init met exp\ne.g. Removed"
    ft_peptide: NotRequired[str]
    "Ft peptide\ne.g. Removed"
    ft_peptide_exp: NotRequired[str]
    "Ft peptide exp\ne.g. Removed"
    ft_signal: NotRequired[str]
    "Ft signal\ne.g. cleaved"
    ft_signal_exp: NotRequired[str]
    "Ft signal exp\ne.g. cleaved"
    ft_propep: NotRequired[str]
    "Ft propep\ne.g. Activation peptide"
    ft_propep_exp: NotRequired[str]
    "Ft propep exp\ne.g. Activation peptide"
    ft_transit: NotRequired[str]
    "Ft transit\ne.g. Mitochondrion"
    ft_transit_exp: NotRequired[str]
    "Ft transit exp\ne.g. Mitochondrion"
    cc_developmental_stage: NotRequired[str]
    "Cc developmental stage\ne.g. brain"
    cc_developmental_stage_exp: NotRequired[str]
    "Cc developmental stage exp\ne.g. brain"
    cc_induction: NotRequired[str]
    "Cc induction\ne.g. calcium"
    cc_induction_exp: NotRequired[str]
    "Cc induction exp\ne.g. calcium"
    cc_tissue_specificity: NotRequired[str]
    "Cc tissue specificity\ne.g. pancreas"
    cc_tissue_specificity_exp: NotRequired[str]
    "Cc tissue specificity exp\ne.g. pancreas"
    interactor: NotRequired[str]
    "Binary Interaction\ne.g. EBI-1042898"
    cc_subunit: NotRequired[str]
    "Cc subunit\ne.g. homodimer"
    cc_subunit_exp: NotRequired[str]
    "Cc subunit exp\ne.g. homodimer"
    structure_3d: NotRequired[bool]
    "3D Structure\ne.g. true"
    ft_secstruct: NotRequired[str]
    "Ft secstruct\ne.g. *"
    ft_secstruct_exp: NotRequired[str]
    "Ft secstruct exp\ne.g. *"
    ft_helix: NotRequired[str]
    "Ft helix\ne.g. *"
    ft_helix_exp: NotRequired[str]
    "Ft helix exp\ne.g. *"
    ft_turn: NotRequired[str]
    "Ft turn\ne.g. *"
    ft_turn_exp: NotRequired[str]
    "Ft turn exp\ne.g. *"
    ft_strand: NotRequired[str]
    "Ft strand\ne.g. *"
    ft_strand_exp: NotRequired[str]
    "Ft strand exp\ne.g. *"
    mass: NotRequired[
        tuple[
            Union[
                int,
                Literal["*",],
            ],
            Union[
                int,
                Literal["*",],
            ],
        ]
    ]
    "Mass(Da)\ne.g. [441126 TO 441126]"
    length: NotRequired[
        tuple[
            Union[
                int,
                Literal["*",],
            ],
            Union[
                int,
                Literal["*",],
            ],
        ]
    ]
    "Sequence length\ne.g. [441 TO 450]"
    cc_ap: NotRequired[str]
    "Cc ap\ne.g. tissues"
    cc_ap_exp: NotRequired[str]
    "Cc ap exp\ne.g. tissues"
    cc_ap_apu: NotRequired[str]
    "Cc ap apu\ne.g. *"
    cc_ap_apu_exp: NotRequired[str]
    "Cc ap apu exp\ne.g. *"
    cc_ap_as: NotRequired[str]
    "Cc ap as\ne.g. experimental"
    cc_ap_as_exp: NotRequired[str]
    "Cc ap as exp\ne.g. experimental"
    cc_ap_ai: NotRequired[str]
    "Cc ap ai\ne.g. acetylalanine"
    cc_ap_ai_exp: NotRequired[str]
    "Cc ap ai exp\ne.g. acetylalanine"
    cc_ap_rf: NotRequired[str]
    "Cc ap rf\ne.g. translation"
    cc_ap_rf_exp: NotRequired[str]
    "Cc ap rf exp\ne.g. translation"
    cc_sequence_caution: NotRequired[str]
    "Cc sequence caution\ne.g. translated"
    cc_sequence_caution_exp: NotRequired[str]
    "Cc sequence caution exp\ne.g. translated"
    cc_sc_framesh: NotRequired[str]
    "Frameshift\ne.g. *"
    cc_sc_einit: NotRequired[str]
    "Erroneous initiation\ne.g. extended"
    cc_sc_eterm: NotRequired[str]
    "Erroneous termination\ne.g. translated"
    cc_sc_epred: NotRequired[str]
    "Erroneous gene model prediction\ne.g. *"
    cc_sc_etran: NotRequired[str]
    "Erroneous translation\ne.g. choice"
    cc_sc_misc: NotRequired[str]
    "Cc sc misc\ne.g. sequence"
    cc_sc_misc_exp: NotRequired[str]
    "Cc sc misc exp\ne.g. sequence"
    cc_mass_spectrometry: NotRequired[str]
    "Cc mass spectrometry\ne.g. electrospray"
    cc_mass_spectrometry_exp: NotRequired[str]
    "Cc mass spectrometry exp\ne.g. electrospray"
    cc_polymorphism: NotRequired[str]
    "Cc polymorphism\ne.g. transcript"
    cc_polymorphism_exp: NotRequired[str]
    "Cc polymorphism exp\ne.g. transcript"
    cc_rna_editing: NotRequired[str]
    "Cc rna editing\ne.g. target"
    cc_rna_editing_exp: NotRequired[str]
    "Cc rna editing exp\ne.g. target"
    ft_variants: NotRequired[str]
    "Ft variants\ne.g. colorectal"
    ft_variants_exp: NotRequired[str]
    "Ft variants exp\ne.g. colorectal"
    ft_variant: NotRequired[str]
    "Ft variant\ne.g. colorectal"
    ft_variant_exp: NotRequired[str]
    "Ft variant exp\ne.g. colorectal"
    ft_var_seq: NotRequired[str]
    "Ft var seq\ne.g. isoform"
    ft_var_seq_exp: NotRequired[str]
    "Ft var seq exp\ne.g. isoform"
    ft_non_std: NotRequired[str]
    "Ft non std\ne.g. selenocysteine"
    ft_non_std_exp: NotRequired[str]
    "Ft non std exp\ne.g. selenocysteine"
    ft_non_ter: NotRequired[str]
    "Ft non ter\ne.g. *"
    ft_non_ter_exp: NotRequired[str]
    "Ft non ter exp\ne.g. *"
    ft_non_cons: NotRequired[str]
    "Ft non cons\ne.g. *"
    ft_non_cons_exp: NotRequired[str]
    "Ft non cons exp\ne.g. *"
    ft_conflict: NotRequired[str]
    "Ft conflict\ne.g. *"
    ft_conflict_exp: NotRequired[str]
    "Ft conflict exp\ne.g. *"
    ft_unsure: NotRequired[str]
    "Ft unsure\ne.g. *"
    ft_unsure_exp: NotRequired[str]
    "Ft unsure exp\ne.g. *"
    ft_positional: NotRequired[str]
    "Ft positional\ne.g. colorectal"
    ft_positional_exp: NotRequired[str]
    "Ft positional exp\ne.g. colorectal"
    fragment: NotRequired[bool]
    "Fragment\ne.g. true"
    organelle: NotRequired[Organelle]
    "Encoded in\ne.g. mitochondrion\n* mitochondrion: Mitochondrion\n* plasmid: Plasmid\n* plastid: Plastid\n* chloroplast: Chloroplast\n* cyanelle: Cyanelle\n* apicoplast: Apicoplast\n* organellar chromatophore: Organellar chromatophore\n* non-photosynthetic plastid: Non-photosynthetic plastid\n* nucleomorph: Nucleomorph\n* hydrogenosome: Hydrogenosome"
    precursor: NotRequired[bool]
    "Precursor\ne.g. true"
    tissue: NotRequired[str]
    "Tissue\ne.g. head"
    strain: NotRequired[str]
    "Strain\ne.g. berkeley"
    plasmid: NotRequired[str]
    "Plasmid\ne.g. pO113"
    transposon: NotRequired[str]
    "Transposon\ne.g. tn3"
    ft_domain: NotRequired[str]
    "Ft domain\ne.g. phosphatase"
    ft_domain_exp: NotRequired[str]
    "Ft domain exp\ne.g. phosphatase"
    cc_domain: NotRequired[str]
    "Cc domain\ne.g. conformation"
    cc_domain_exp: NotRequired[str]
    "Cc domain exp\ne.g. conformation"
    family: NotRequired[str]
    "Protein family\ne.g. pa28"
    ft_coiled: NotRequired[str]
    "Ft coiled\ne.g. *"
    ft_coiled_exp: NotRequired[str]
    "Ft coiled exp\ne.g. *"
    ft_compbias: NotRequired[str]
    "Ft compbias\ne.g. basic residues"
    ft_compbias_exp: NotRequired[str]
    "Ft compbias exp\ne.g. basic residues"
    ft_motif: NotRequired[str]
    "Ft motif\ne.g. motif"
    ft_motif_exp: NotRequired[str]
    "Ft motif exp\ne.g. motif"
    ft_region: NotRequired[str]
    "Ft region\ne.g. motif"
    ft_region_exp: NotRequired[str]
    "Ft region exp\ne.g. motif"
    ft_repeat: NotRequired[str]
    "Ft repeat\ne.g. motif"
    ft_repeat_exp: NotRequired[str]
    "Ft repeat exp\ne.g. motif"
    cc_similarity: NotRequired[str]
    "Cc similarity\ne.g. phosphatase"
    cc_similarity_exp: NotRequired[str]
    "Cc similarity exp\ne.g. phosphatase"
    ft_zn_fing: NotRequired[str]
    "Ft zn fing\ne.g. UBP"
    ft_zn_fing_exp: NotRequired[str]
    "Ft zn fing exp\ne.g. UBP"
    xref: NotRequired[str]
    "Any cross-reference"
    database: NotRequired[str]
    "Database\ne.g. Bgee"
    cc_webresource: NotRequired[str]
    "Cc webresource\ne.g. lck"
    cc_webresource_exp: NotRequired[str]
    "Cc webresource exp\ne.g. lck"
    date_created: NotRequired[
        tuple[
            Union[
                date,
                Literal["*",],
            ],
            Union[
                date,
                Literal["*",],
            ],
        ]
    ]
    "Date Of Creation\ne.g. [2018-03-04 TO 2018-03-08]"
    date_modified: NotRequired[
        tuple[
            Union[
                date,
                Literal["*",],
            ],
            Union[
                date,
                Literal["*",],
            ],
        ]
    ]
    "Date of last entry modification\ne.g. [2018-03-04 TO 2018-03-08]"
    date_sequence_modified: NotRequired[
        tuple[
            Union[
                date,
                Literal["*",],
            ],
            Union[
                date,
                Literal["*",],
            ],
        ]
    ]
    "Date of last sequence modification\ne.g. [2018-03-04 TO 2018-03-08]"
    go: NotRequired[str]
    "Go\ne.g. 0009986"
    go_manual: NotRequired[str]
    "go, any manual assertion"
    go_automatic: NotRequired[str]
    "go, any automatic assertion"
    go_exp: NotRequired[str]
    "go, inferred from experiment [exp]"
    go_iba: NotRequired[str]
    "go, inferred from biological aspect of ancestor [iba]"
    go_ic: NotRequired[str]
    "go, inferred by curator [ic]"
    go_ida: NotRequired[str]
    "go, inferred from direct assay [ida]"
    go_iep: NotRequired[str]
    "go, inferred from expression pattern [iep]"
    go_igc: NotRequired[str]
    "go, inferred from genomic context [igc]"
    go_igi: NotRequired[str]
    "go, inferred from genetic interaction [igi]"
    go_imp: NotRequired[str]
    "go, inferred from mutant phenotype [imp]"
    go_ipi: NotRequired[str]
    "go, inferred from physical interaction [ipi]"
    go_isa: NotRequired[str]
    "go, inferred from sequence alignment [isa]"
    go_ism: NotRequired[str]
    "go, inferred from sequence mode [ism]"
    go_iso: NotRequired[str]
    "go, inferred from sequence orthology [iso]"
    go_iss: NotRequired[str]
    "go, inferred from sequence or structural similarity [iss]"
    go_nas: NotRequired[str]
    "go, non-traceable author statement [nas]"
    go_tas: NotRequired[str]
    "go, traceable author statement [tas]"
    go_hda: NotRequired[str]
    "go, inferred from high throughput direct assay [hda]"
    go_hmp: NotRequired[str]
    "go, inferred from high throughput mutant phenotype [hmp]"
    go_hgi: NotRequired[str]
    "go, inferred from high throughput genetic interaction [hgi]"
    go_hep: NotRequired[str]
    "go, interred from high throughput expression pattern [hep]"
    go_htp: NotRequired[str]
    "go, inferred from high throughput experiment [htp]"
    go_iea: NotRequired[str]
    "go, inferred from electronic annotation [iea]"
    chebi: NotRequired[str]
    "CHEBI ID\ne.g. 29105"
    inchikey: NotRequired[str]
    "InChIKey\ne.g. XLYOFNOQVPJJNP-UHFFFAOYSA-N"
    keyword: NotRequired[str]
    "Keyword [KW]\ne.g. chromosomal"
    lit_author: NotRequired[str]
    "Author\ne.g. smith"
    lit_journal: NotRequired[str]
    "Journal\ne.g. nature"
    lit_pubdate: NotRequired[
        tuple[
            Union[
                date,
                Literal["*",],
            ],
            Union[
                date,
                Literal["*",],
            ],
        ]
    ]
    "Published\ne.g. [2009-01-01 TO 2019-12-31]"
    lit_pubmed: NotRequired[str]
    "PubMed ID\ne.g. 15165820"
    lit_title: NotRequired[str]
    "Title\ne.g. protein"
    lit_citation_id: NotRequired[str]
    "Citation ID\ne.g. CI-6EPRJ6MFFS5LC"
    computational_pubmed_id: NotRequired[str]
    "Computational PubMed ID\ne.g. 15165820"
    community_pubmed_id: NotRequired[str]
    "Community PubMed ID\ne.g. 15165820"
    proteome: NotRequired[str]
    "Proteome ID\ne.g. UP000005640"
    proteomecomponent: NotRequired[str]
    "Proteome Component\ne.g. chromosome"
    scope: NotRequired[str]
    "Cited for\ne.g. microtubule"
    reviewed: NotRequired[bool]
    "Reviewed\ne.g. true"
    active: NotRequired[bool]
    "Active\ne.g. true"
    uniref_cluster_50: NotRequired[str]
    "UniRef50\ne.g. UniRef50_P05067"
    uniref_cluster_90: NotRequired[str]
    "UniRef90\ne.g. UniRef90_P05067"
    uniref_cluster_100: NotRequired[str]
    "UniRef100\ne.g. UniRef100_P05067"
    uniparc: NotRequired[str]
    "UniParc ID\ne.g. UPI000002DB1C"


UniprotkbQuery: TypeAlias = Union[UniprotkbQueryDict, str]
UniprotkbNamesTaxonomy: TypeAlias = Literal[
    "accession",
    "id",
    "gene_names",
    "gene_oln",
    "gene_orf",
    "gene_primary",
    "gene_synonym",
    "organism_name",
    "organism_id",
    "protein_name",
    "xref_proteomes",
    "lineage",
    "lineage_ids",
    "virus_hosts",
]
UniprotkbSequences: TypeAlias = Literal[
    "cc_alternative_products",
    "ft_var_seq",
    "error_gmodel_pred",
    "fragment",
    "organelle",
    "length",
    "mass",
    "cc_mass_spectrometry",
    "ft_variant",
    "ft_non_cons",
    "ft_non_std",
    "ft_non_ter",
    "cc_polymorphism",
    "cc_rna_editing",
    "sequence",
    "cc_sequence_caution",
    "ft_conflict",
    "ft_unsure",
    "sequence_version",
]
UniprotkbFunction: TypeAlias = Literal[
    "absorption",
    "ft_act_site",
    "ft_binding",
    "cc_catalytic_activity",
    "cc_cofactor",
    "ft_dna_bind",
    "ec",
    "cc_activity_regulation",
    "cc_function",
    "kinetics",
    "cc_pathway",
    "ph_dependence",
    "redox_potential",
    "rhea",
    "ft_site",
    "temp_dependence",
]
UniprotkbMiscellaneous: TypeAlias = Literal[
    "annotation_score",
    "cc_caution",
    "keyword",
    "keywordid",
    "cc_miscellaneous",
    "protein_existence",
    "reviewed",
    "tools",
    "uniparc_id",
    "comment_count",
    "feature_count",
]
UniprotkbInteraction: TypeAlias = Literal["cc_interaction", "cc_subunit"]
UniprotkbExpression: TypeAlias = Literal[
    "cc_developmental_stage", "cc_induction", "cc_tissue_specificity"
]
UniprotkbGeneOntologyGo: TypeAlias = Literal["go_p", "go_c", "go", "go_f", "go_id"]
UniprotkbPathologyBiotech: TypeAlias = Literal[
    "cc_allergen",
    "cc_biotechnology",
    "cc_disruption_phenotype",
    "cc_disease",
    "ft_mutagen",
    "cc_pharmaceutical",
    "cc_toxic_dose",
]
UniprotkbSubcellularLocation: TypeAlias = Literal[
    "ft_intramem", "cc_subcellular_location", "ft_topo_dom", "ft_transmem"
]
UniprotkbPtmProcessing: TypeAlias = Literal[
    "ft_chain",
    "ft_crosslnk",
    "ft_disulfid",
    "ft_carbohyd",
    "ft_init_met",
    "ft_lipid",
    "ft_mod_res",
    "ft_peptide",
    "cc_ptm",
    "ft_propep",
    "ft_signal",
    "ft_transit",
]
UniprotkbStructure: TypeAlias = Literal[
    "structure_3d", "ft_strand", "ft_helix", "ft_turn"
]
UniprotkbPublications: TypeAlias = Literal["lit_pubmed_id", "lit_doi_id"]
UniprotkbDateOf: TypeAlias = Literal[
    "date_created", "date_modified", "date_sequence_modified", "version"
]
UniprotkbFamilyDomains: TypeAlias = Literal[
    "ft_coiled",
    "ft_compbias",
    "cc_domain",
    "ft_domain",
    "ft_motif",
    "protein_families",
    "ft_region",
    "ft_repeat",
    "cc_similarity",
    "ft_zn_fing",
]
UniprotkbSequence: TypeAlias = Literal[
    "xref_ccds", "xref_embl", "xref_generif", "xref_pir", "xref_refseq"
]
UniprotkbThreedStructure: TypeAlias = Literal[
    "xref_alphafolddb",
    "xref_bmrb",
    "xref_emdb",
    "xref_pcddb",
    "xref_pdb",
    "xref_pdbsum",
    "xref_sasbdb",
    "xref_smr",
]
UniprotkbProteinProteinInteraction: TypeAlias = Literal[
    "xref_biogrid",
    "xref_corum",
    "xref_complexportal",
    "xref_dip",
    "xref_elm",
    "xref_intact",
    "xref_mint",
    "xref_string",
]
UniprotkbChemistry: TypeAlias = Literal[
    "xref_bindingdb",
    "xref_chembl",
    "xref_drugbank",
    "xref_drugcentral",
    "xref_guidetopharmacology",
    "xref_swisslipids",
]
UniprotkbProteinFamilygroup: TypeAlias = Literal[
    "xref_allergome",
    "xref_cazy",
    "xref_esther",
    "xref_imgt_gene-db",
    "xref_merops",
    "xref_moondb",
    "xref_moonprot",
    "xref_peroxibase",
    "xref_rebase",
    "xref_tcdb",
    "xref_unilectin",
]
UniprotkbPtm: TypeAlias = Literal[
    "xref_carbonyldb",
    "xref_depod",
    "xref_glyconnect",
    "xref_glycosmos",
    "xref_glygen",
    "xref_metosite",
    "xref_phosphositeplus",
    "xref_swisspalm",
    "xref_unicarbkb",
    "xref_iptmnet",
]
UniprotkbPolymorphismAndMutation: TypeAlias = Literal[
    "xref_alzforum", "xref_biomuta", "xref_dmdm", "xref_dbsnp"
]
UniprotkbTwodGel: TypeAlias = Literal[
    "xref_compluyeast-2dpage", "xref_ogp", "xref_reproduction-2dpage"
]
UniprotkbProteomic: TypeAlias = Literal[
    "xref_cptac",
    "xref_massive",
    "xref_pride",
    "xref_paxdb",
    "xref_peptideatlas",
    "xref_promex",
    "xref_proteomicsdb",
    "xref_pumba",
    "xref_topdownproteomics",
    "xref_jpost",
]
UniprotkbProtocolsAndMaterials: TypeAlias = Literal[
    "xref_abcd", "xref_antibodypedia", "xref_cptc", "xref_dnasu"
]
UniprotkbGenomeAnnotation: TypeAlias = Literal[
    "xref_ensembl",
    "xref_ensemblbacteria",
    "xref_ensemblfungi",
    "xref_ensemblmetazoa",
    "xref_ensemblplants",
    "xref_ensemblprotists",
    "xref_geneid",
    "xref_gramene",
    "xref_kegg",
    "xref_mane-select",
    "xref_patric",
    "xref_ucsc",
    "xref_vectorbase",
    "xref_wbparasite",
]
UniprotkbOrganismSpecific: TypeAlias = Literal[
    "xref_agr",
    "xref_arachnoserver",
    "xref_araport",
    "xref_cgd",
    "xref_ctd",
    "xref_conoserver",
    "xref_disgenet",
    "xref_echobase",
    "xref_flybase",
    "xref_genecards",
    "xref_genereviews",
    "xref_hgnc",
    "xref_hpa",
    "xref_ic4r",
    "xref_japonicusdb",
    "xref_legiolist",
    "xref_leproma",
    "xref_mgi",
    "xref_mim",
    "xref_maizegdb",
    "xref_malacards",
    "xref_niagads",
    "xref_opentargets",
    "xref_orphanet",
    "xref_pharmgkb",
    "xref_pombase",
    "xref_pseudocap",
    "xref_rgd",
    "xref_sgd",
    "xref_tair",
    "xref_tuberculist",
    "xref_veupathdb",
    "xref_vgnc",
    "xref_wormbase",
    "xref_xenbase",
    "xref_zfin",
    "xref_dictybase",
    "xref_euhcvdb",
    "xref_nextprot",
]
UniprotkbPhylogenomic: TypeAlias = Literal[
    "xref_genetree",
    "xref_hogenom",
    "xref_inparanoid",
    "xref_oma",
    "xref_orthodb",
    "xref_phylomedb",
    "xref_treefam",
    "xref_eggnog",
]
UniprotkbEnzymeAndPathway: TypeAlias = Literal[
    "xref_brenda",
    "xref_biocyc",
    "xref_pathwaycommons",
    "xref_plantreactome",
    "xref_reactome",
    "xref_sabio-rk",
    "xref_signor",
    "xref_signalink",
    "xref_unipathway",
]
UniprotkbOther: TypeAlias = Literal[
    "xref_biogrid-orcs",
    "xref_chitars",
    "xref_evolutionarytrace",
    "xref_genewiki",
    "xref_genomernai",
    "xref_orcid",
    "xref_pgenn",
    "xref_phi-base",
    "xref_pro",
    "xref_pharos",
    "xref_pubtator",
    "xref_rnact",
    "xref_emind",
]
UniprotkbGeneExpression: TypeAlias = Literal[
    "xref_bgee", "xref_cleanex", "xref_collectf", "xref_expressionatlas"
]
UniprotkbFamilyAndDomain: TypeAlias = Literal[
    "xref_cdd",
    "xref_disprot",
    "xref_gene3d",
    "xref_hamap",
    "xref_ideal",
    "xref_interpro",
    "xref_ncbifam",
    "xref_panther",
    "xref_pirsf",
    "xref_prints",
    "xref_prosite",
    "xref_pfam",
    "xref_sfld",
    "xref_smart",
    "xref_supfam",
]
UniprotkbFields: TypeAlias = Literal[
    UniprotkbNamesTaxonomy,
    UniprotkbSequences,
    UniprotkbFunction,
    UniprotkbMiscellaneous,
    UniprotkbInteraction,
    UniprotkbExpression,
    UniprotkbGeneOntologyGo,
    UniprotkbPathologyBiotech,
    UniprotkbSubcellularLocation,
    UniprotkbPtmProcessing,
    UniprotkbStructure,
    UniprotkbPublications,
    UniprotkbDateOf,
    UniprotkbFamilyDomains,
    UniprotkbSequence,
    UniprotkbThreedStructure,
    UniprotkbProteinProteinInteraction,
    UniprotkbChemistry,
    UniprotkbProteinFamilygroup,
    UniprotkbPtm,
    UniprotkbPolymorphismAndMutation,
    UniprotkbTwodGel,
    UniprotkbProteomic,
    UniprotkbProtocolsAndMaterials,
    UniprotkbGenomeAnnotation,
    UniprotkbOrganismSpecific,
    UniprotkbPhylogenomic,
    UniprotkbEnzymeAndPathway,
    UniprotkbOther,
    UniprotkbGeneExpression,
    UniprotkbFamilyAndDomain,
]



from __future__ import annotations

from datetime import date
from typing import Iterable, Union

from typing_extensions import Literal, NotRequired, TypeAlias, TypedDict

ProteomeType: TypeAlias = Literal["1", "2", "3", "4"]
Cpd: TypeAlias = Literal["1", "2", "3", "4", "5", "6"]


class ProteomesQueryDict(TypedDict):
    and_: NotRequired[Iterable["ProteomesQuery"]]
    "Two or more filters that must both be satisfied"
    or_: NotRequired[Iterable["ProteomesQuery"]]
    "Two or more filters, any of which can be satisfied"
    not_: NotRequired[Iterable["ProteomesQuery"]]
    "Negate a filter"
    upid: NotRequired[str]
    "Proteome ID\ne.g. UP000000718"
    proteome_type: NotRequired[ProteomeType]
    "Proteome Type\ne.g. 1\n* 1: Reference\n* 2: Other\n* 3: Redundant\n* 4: Excluded"
    organism_name: NotRequired[str]
    "Organism [OS]\ne.g. Human"
    organism_id: NotRequired[str]
    "Organism id"
    taxonomy_name: NotRequired[str]
    "Taxonomy [OC]\ne.g. Human"
    taxonomy_id: NotRequired[str]
    "Taxonomy id"
    genome_accession: NotRequired[str]
    "Genome Accession\ne.g. CM000663"
    genome_assembly: NotRequired[str]
    "Genome Assembly\ne.g. GCA_000001405.27"
    cpd: NotRequired[Cpd]
    "CPD (Complete Proteome Detector)\ne.g. 1\n* 1: Standard\n* 2: Close to standard (high value)\n* 3: Close to standard (low value)\n* 4: Outlier (high value)\n* 5: Outlier (low value)\n* 6: Unknown"
    busco: NotRequired[
        tuple[
            Union[
                int,
                Literal["*",],
            ],
            Union[
                int,
                Literal["*",],
            ],
        ]
    ]
    "BUSCO (Complete %)\ne.g. 97"


ProteomesQuery: TypeAlias = Union[ProteomesQueryDict, str]
ProteomesNamesTaxonomy: TypeAlias = Literal[
    "upid", "organism", "organism_id", "components", "mnemonic", "lineage"
]
ProteomesMiscellaneous: TypeAlias = Literal[
    "busco", "cpd", "genome_assembly", "genome_representation", "protein_count"
]
ProteomesFields: TypeAlias = Literal[ProteomesNamesTaxonomy, ProteomesMiscellaneous]



from __future__ import annotations

from datetime import date
from typing import Iterable, Union

from typing_extensions import Literal, NotRequired, TypeAlias, TypedDict

Database: TypeAlias = Literal[
    "EnsemblBacteria",
    "EnsemblFungi",
    "EnsemblMetazoa",
    "EnsemblPlants",
    "EnsemblProtists",
    "embl-cds",
    "EMBL_CON",
    "EMBL_TPA",
    "EMBL_TSA",
    "EMBLWGS",
    "Ensembl",
    "EnsemblRapid",
    "EPO",
    "FlyBase",
    "FusionGDB",
    "H-InvDB",
    "IPI",
    "JPO",
    "KIPO",
    "PATRIC",
    "PDB",
    "PIR",
    "PIRARC",
    "PRF",
    "RefSeq",
    "REMTREMBL",
    "SEED",
    "SGD",
    "UniProt",
    "isoforms",
    "TAIR",
    "TREMBLNEW",
    "TREMBL_VARSPLIC",
    "TROME",
    "UNIMES",
    "USPTO",
    "VectorBase",
    "VEGA",
    "WBParaSite",
    "WormBase",
]
Active: TypeAlias = Literal[
    "EnsemblBacteria",
    "EnsemblFungi",
    "EnsemblMetazoa",
    "EnsemblPlants",
    "EnsemblProtists",
    "embl-cds",
    "EMBL_CON",
    "EMBL_TSA",
    "EMBLWGS",
    "Ensembl",
    "EnsemblRapid",
    "EPO",
    "FlyBase",
    "FusionGDB",
    "JPO",
    "KIPO",
    "PATRIC",
    "PDB",
    "RefSeq",
    "SEED",
    "SGD",
    "UniProt",
    "isoforms",
    "TAIR",
    "TROME",
    "USPTO",
    "VEGA",
    "WBParaSite",
    "WormBase",
]


class UniparcQueryDict(TypedDict):
    and_: NotRequired[Iterable["UniparcQuery"]]
    "Two or more filters that must both be satisfied"
    or_: NotRequired[Iterable["UniparcQuery"]]
    "Two or more filters, any of which can be satisfied"
    not_: NotRequired[Iterable["UniparcQuery"]]
    "Negate a filter"
    upi: NotRequired[str]
    "UniParc ID\ne.g. UPI0000000001"
    uniprotkb: NotRequired[str]
    "UniProtKB AC\ne.g. P12345"
    isoform: NotRequired[str]
    "UniProtKB isoform ID\ne.g. P12345"
    upid: NotRequired[str]
    "Proteome ID\ne.g. UP123456789"
    taxonomy_name: NotRequired[str]
    "Taxonomy [OC]\ne.g. Human"
    taxonomy_id: NotRequired[str]
    "Taxonomy id"
    gene: NotRequired[str]
    "Gene name [GN]\ne.g. PROZ"
    protein: NotRequired[str]
    "Protein name\ne.g. Protein Z"
    database: NotRequired[Database]
    "Database\ne.g. Gene3D\n* EnsemblBacteria: EnsemblBacteria\n* EnsemblFungi: EnsemblFungi\n* EnsemblMetazoa: EnsemblMetazoa\n* EnsemblPlants: EnsemblPlants\n* EnsemblProtists: EnsemblProtists\n* embl-cds: EMBL CDS\n* EMBL_CON: EMBL_CON\n* EMBL_TPA: EMBL_TPA\n* EMBL_TSA: EMBL_TSA\n* EMBLWGS: EMBLWGS\n* Ensembl: Ensembl\n* EnsemblRapid: EnsemblRapid\n* EPO: EPO\n* FlyBase: FlyBase\n* FusionGDB: FusionGDB\n* H-InvDB: H-InvDB\n* IPI: IPI\n* JPO: JPO\n* KIPO: KIPO\n* PATRIC: PATRIC\n* PDB: PDB\n* PIR: PIR\n* PIRARC: PIRARC\n* PRF: PRF\n* RefSeq: RefSeq\n* REMTREMBL: REMTREMBL\n* SEED: SEED\n* SGD: SGD\n* UniProt: UniProtKB\n* isoforms: UniProtKB/Swiss-Prot isoforms\n* TAIR: TAIR\n* TREMBLNEW: TREMBLNEW\n* TREMBL_VARSPLIC: TREMBL_VARSPLIC\n* TROME: TROME\n* UNIMES: UNIMES\n* USPTO: USPTO\n* VectorBase: VectorBase\n* VEGA: VEGA\n* WBParaSite: WBParaSite\n* WormBase: WormBase"
    active: NotRequired[Active]
    "Active\ne.g. Gene3D\n* EnsemblBacteria: EnsemblBacteria\n* EnsemblFungi: EnsemblFungi\n* EnsemblMetazoa: EnsemblMetazoa\n* EnsemblPlants: EnsemblPlants\n* EnsemblProtists: EnsemblProtists\n* embl-cds: EMBL CDS\n* EMBL_CON: EMBL_CON\n* EMBL_TSA: EMBL_TSA\n* EMBLWGS: EMBLWGS\n* Ensembl: Ensembl\n* EnsemblRapid: EnsemblRapid\n* EPO: EPO\n* FlyBase: FlyBase\n* FusionGDB: FusionGDB\n* JPO: JPO\n* KIPO: KIPO\n* PATRIC: PATRIC\n* PDB: PDB\n* RefSeq: RefSeq\n* SEED: SEED\n* SGD: SGD\n* UniProt: UniProtKB\n* isoforms: UniProtKB/Swiss-Prot isoforms\n* TAIR: TAIR\n* TROME: TROME\n* USPTO: USPTO\n* VEGA: VEGA\n* WBParaSite: WBParaSite\n* WormBase: WormBase"
    checksum: NotRequired[str]
    "Checksum (CRC64/MD5)\ne.g. B8824CE1ECAEEEAE"
    length: NotRequired[
        tuple[
            Union[
                int,
                Literal["*",],
            ],
            Union[
                int,
                Literal["*",],
            ],
        ]
    ]
    "Sequence length\ne.g. [100 TO 300]"
    dbid: NotRequired[str]
    "Database ID\ne.g. AAC02967"
    feature_id: NotRequired[str]
    "Feature ID\ne.g. IPR004251"
    proteomecomponent: NotRequired[str]
    "Proteome Component\ne.g. chromosome"
    organism_id: NotRequired[int]
    "Organism ID\ne.g. 10254"


UniparcQuery: TypeAlias = Union[UniparcQueryDict, str]
UniparcNamesTaxonomy: TypeAlias = Literal[
    "upi", "gene", "organism_id", "organism", "protein", "proteome"
]
UniparcSequences: TypeAlias = Literal["checksum", "length", "sequence"]
UniparcMiscellaneous: TypeAlias = Literal["accession",]
UniparcDateOf: TypeAlias = Literal["first_seen", "last_seen"]
UniparcFamilyDomains: TypeAlias = Literal[
    "CDD",
    "Gene3D",
    "HAMAP",
    "PANTHER",
    "Pfam",
    "PIRSF",
    "PRINTS",
    "PROSITE",
    "SFLD",
    "SMART",
    "SUPFAM",
    "NCBIfam",
]
UniparcFields: TypeAlias = Literal[
    UniparcNamesTaxonomy,
    UniparcSequences,
    UniparcMiscellaneous,
    UniparcDateOf,
    UniparcFamilyDomains,
]



from typing import Any, Mapping

from typing_extensions import Literal

from unipressed.dataset.core import FetchManyClient
from unipressed.dataset.generated_types.uniprotkb import UniprotkbFields, UniprotkbQuery

UniprotkbFormats = Literal["txt", "xml", "fasta", "gff", "json", "list", "tsv", "xlsx"]


class UniprotkbClient(
    FetchManyClient[
        UniprotkbQuery,
        Mapping[str, Any],
        UniprotkbFields,
        UniprotkbFormats,
    ]
):
    @classmethod
    def name(cls):
        return "uniprotkb"

    @classmethod
    def _bulk_endpoint(cls):
        return "accessions"

    @classmethod
    def _id_field(cls, record: Mapping[str, Any]) -> str:
        return record["primaryAccession"]



from typing import Any, Mapping

from typing_extensions import Literal

from unipressed.dataset.core import DatasetClient
from unipressed.dataset.generated_types.diseases import DiseasesFields, DiseasesQuery

DiseasesFormats = Literal["json", "tsv", "xlsx", "list", "obo"]


class DiseasesClient(
    DatasetClient[
        DiseasesQuery,
        Mapping[str, Any],
        DiseasesFields,
        DiseasesFormats,
    ]
):
    @classmethod
    def name(cls):
        return "diseases"



from typing import Any, Mapping

from typing_extensions import Literal

from unipressed.dataset.core import DatasetClient
from unipressed.dataset.generated_types.locations import LocationsFields, LocationsQuery

LocationsFormats = Literal["json", "tsv", "xlsx", "list", "obo"]


class LocationsClient(
    DatasetClient[
        LocationsQuery,
        Mapping[str, Any],
        LocationsFields,
        LocationsFormats,
    ]
):
    @classmethod
    def name(cls):
        return "locations"



from __future__ import annotations

from datetime import date
from typing import Iterable, Union

from typing_extensions import Literal, NotRequired, TypeAlias, TypedDict


class DiseasesQueryDict(TypedDict):
    and_: NotRequired[Iterable["DiseasesQuery"]]
    "Two or more filters that must both be satisfied"
    or_: NotRequired[Iterable["DiseasesQuery"]]
    "Two or more filters, any of which can be satisfied"
    not_: NotRequired[Iterable["DiseasesQuery"]]
    "Negate a filter"
    name: NotRequired[str]
    "Name\ne.g. alzheimer disease"
    id: NotRequired[str]
    "Disease [AC]\ne.g. DI-12345"


DiseasesQuery: TypeAlias = Union[DiseasesQueryDict, str]
DiseasesDisease: TypeAlias = Literal[
    "id",
    "name",
    "acronym",
    "definition",
    "alternative_names",
    "cross_references",
    "keywords",
    "statistics",
]
DiseasesFields: TypeAlias = Literal[DiseasesDisease,]



from typing import Any, Mapping, TypeVar

QueryType = TypeVar("QueryType")
JsonResultType = TypeVar("JsonResultType", bound=Mapping[str, Any])
FieldsType = TypeVar("FieldsType", bound=str)
FormatType = TypeVar("FormatType", bound=str)



from typing import Any, Mapping

from typing_extensions import Literal

from unipressed.dataset.core import DatasetClient
from unipressed.dataset.generated_types.unirule import UniruleFields, UniruleQuery

UniruleFormats = Literal["json", "list"]


class UniruleClient(
    DatasetClient[UniruleQuery, Mapping[str, Any], UniruleFields, UniruleFormats]
):
    @classmethod
    def name(cls):
        return "unirule"

    @classmethod
    def _id_field(cls, record: Mapping[str, Any]) -> str:
        return record["uniRuleId"]



from __future__ import annotations

from typing import Any, Iterable

import requests


def iter_pages(
    response: requests.Response, method: str = "GET"
) -> Iterable[requests.Response]:
    yield response
    while True:
        link: dict[str, Any] | None = response.links.get("next")
        if link is not None:
            response = requests.request(method, url=link["url"])
            yield response
        else:
            break



from typing import Any, Mapping

from typing_extensions import Literal

from unipressed.dataset.core import DatasetClient
from unipressed.dataset.generated_types.citations import CitationsFields, CitationsQuery

CitationsFormats = Literal["json", "list", "tsv", "xlsx"]


class CitationsClient(
    DatasetClient[
        CitationsQuery,
        Mapping[str, Any],
        CitationsFields,
        CitationsFormats,
    ]
):
    @classmethod
    def name(cls):
        return "citations"

    @classmethod
    def _id_field(cls, record: Any) -> str:
        return record["citation"]["id"]



from typing import Iterable, Optional

from typing_extensions import Literal, overload


class SubmitDummyClass:
    @classmethod
    @overload
    def submit(
        cls,
        source: Literal["UniProtKB_AC-ID",],
        dest: Literal[
            "CCDS",
            "PIR",
            "PDB",
            "BioGRID",
            "ComplexPortal",
            "DIP",
            "STRING",
            "ChEMBL",
            "DrugBank",
            "GuidetoPHARMACOLOGY",
            "SwissLipids",
            "Allergome",
            "ESTHER",
            "MEROPS",
            "PeroxiBase",
            "REBASE",
            "TCDB",
            "GlyConnect",
            "BioMuta",
            "DMDM",
            "CPTAC",
            "ProteomicsDB",
            "DNASU",
            "Ensembl",
            "GeneID",
            "KEGG",
            "PATRIC",
            "UCSC",
            "WBParaSite",
            "ArachnoServer",
            "Araport",
            "CGD",
            "ConoServer",
            "dictyBase",
            "EchoBASE",
            "euHCVdb",
            "VEuPathDB",
            "FlyBase",
            "GeneCards",
            "GeneReviews",
            "HGNC",
            "LegioList",
            "Leproma",
            "MaizeGDB",
            "MGI",
            "MIM",
            "neXtProt",
            "OpenTargets",
            "Orphanet",
            "PharmGKB",
            "PomBase",
            "PseudoCAP",
            "RGD",
            "SGD",
            "TubercuList",
            "VGNC",
            "WormBase",
            "Xenbase",
            "ZFIN",
            "eggNOG",
            "GeneTree",
            "HOGENOM",
            "OMA",
            "OrthoDB",
            "TreeFam",
            "BioCyc",
            "Reactome",
            "UniPathway",
            "PlantReactome",
            "ChiTaRS",
            "GeneWiki",
            "GenomeRNAi",
            "PHI-base",
            "CollecTF",
            "IDEAL",
            "DisProt",
            "UniProtKB",
            "UniProtKB-Swiss-Prot",
            "UniParc",
            "UniRef50",
            "UniRef90",
            "UniRef100",
            "Gene_Name",
            "CRC64",
            "EMBL-GenBank-DDBJ",
            "EMBL-GenBank-DDBJ_CDS",
            "GI_number",
            "RefSeq_Nucleotide",
            "RefSeq_Protein",
            "Ensembl_Protein",
            "Ensembl_Transcript",
            "Ensembl_Genomes",
            "Ensembl_Genomes_Protein",
            "Ensembl_Genomes_Transcript",
            "WBParaSite_Transcript-Protein",
            "WormBase_Protein",
            "WormBase_Transcript",
        ],
        ids: Iterable[str],
    ):
        ...

    @classmethod
    @overload
    def submit(
        cls,
        source: Literal["UniParc",],
        dest: Literal["UniProtKB", "UniProtKB-Swiss-Prot", "UniParc"],
        ids: Iterable[str],
    ):
        ...

    @classmethod
    @overload
    def submit(
        cls,
        source: Literal["UniRef50",],
        dest: Literal["UniProtKB", "UniProtKB-Swiss-Prot", "UniRef50"],
        ids: Iterable[str],
    ):
        ...

    @classmethod
    @overload
    def submit(
        cls,
        source: Literal["UniRef90",],
        dest: Literal["UniProtKB", "UniProtKB-Swiss-Prot", "UniRef90"],
        ids: Iterable[str],
    ):
        ...

    @classmethod
    @overload
    def submit(
        cls,
        source: Literal["UniRef100",],
        dest: Literal["UniProtKB", "UniProtKB-Swiss-Prot", "UniRef100"],
        ids: Iterable[str],
    ):
        ...

    @classmethod
    @overload
    def submit(
        cls,
        source: Literal["Gene_Name",],
        dest: Literal["UniProtKB", "UniProtKB-Swiss-Prot"],
        ids: Iterable[str],
        taxon_id: Optional[int] = None,
    ):
        ...

    @classmethod
    @overload
    def submit(
        cls,
        source: Literal[
            "CRC64",
            "CCDS",
            "EMBL-GenBank-DDBJ",
            "EMBL-GenBank-DDBJ_CDS",
            "GI_number",
            "PIR",
            "RefSeq_Nucleotide",
            "RefSeq_Protein",
            "PDB",
            "BioGRID",
            "ComplexPortal",
            "DIP",
            "STRING",
            "ChEMBL",
            "DrugBank",
            "GuidetoPHARMACOLOGY",
            "SwissLipids",
            "Allergome",
            "ESTHER",
            "MEROPS",
            "PeroxiBase",
            "REBASE",
            "TCDB",
            "GlyConnect",
            "BioMuta",
            "DMDM",
            "CPTAC",
            "ProteomicsDB",
            "DNASU",
            "Ensembl",
            "Ensembl_Genomes",
            "Ensembl_Genomes_Protein",
            "Ensembl_Genomes_Transcript",
            "Ensembl_Protein",
            "Ensembl_Transcript",
            "GeneID",
            "KEGG",
            "PATRIC",
            "UCSC",
            "WBParaSite",
            "WBParaSite_Transcript-Protein",
            "ArachnoServer",
            "Araport",
            "CGD",
            "ConoServer",
            "dictyBase",
            "EchoBASE",
            "euHCVdb",
            "FlyBase",
            "GeneCards",
            "GeneReviews",
            "HGNC",
            "LegioList",
            "Leproma",
            "MaizeGDB",
            "MGI",
            "MIM",
            "neXtProt",
            "OpenTargets",
            "Orphanet",
            "PharmGKB",
            "PomBase",
            "PseudoCAP",
            "RGD",
            "SGD",
            "TubercuList",
            "VEuPathDB",
            "VGNC",
            "WormBase",
            "WormBase_Protein",
            "WormBase_Transcript",
            "Xenbase",
            "ZFIN",
            "eggNOG",
            "GeneTree",
            "HOGENOM",
            "OMA",
            "OrthoDB",
            "TreeFam",
            "BioCyc",
            "PlantReactome",
            "Reactome",
            "UniPathway",
            "ChiTaRS",
            "GeneWiki",
            "GenomeRNAi",
            "PHI-base",
            "CollecTF",
            "DisProt",
            "IDEAL",
        ],
        dest: Literal["UniProtKB", "UniProtKB-Swiss-Prot"],
        ids: Iterable[str],
    ):
        ...

    @classmethod
    def submit(
        cls, source: str, dest: str, ids: Iterable[str], taxon_id: Optional[int] = None
    ):
        ...
